from mlhandmade.kernel_methods.kernel import Kernel
from mlhandmade.kernel_methods.rbf import RBFKernel
from mlhandmade.kernel_methods.poly import PolynomialKernel
from mlhandmade.kernel_methods.linear import LinearKernel
from mlhandmade.kernel_methods.svm import (
    SupportVectorClassifier,
    SupportVectorRegressor
)
from mlhandmade.kernel_methods.kernel_ridge import KernelRidge
__all__ = [
    "Kernel",
    "RBFKernel",
    "PolynomialKernel",
    "LinearKernel",
    "SupportVectorClassifier",
    "SupportVectorRegressor",
    "KernelRidge"
]


import numpy as np

from ..base import BaseEstimator
from .linear import LinearKernel
from .poly import PolynomialKernel
from .rbf import RBFKernel

kernels = {
    "linear" : LinearKernel,
    "poly" : PolynomialKernel,
    "rbf" : RBFKernel
}

class KernelRidge(BaseEstimator):
    """
    Implementation of kernel ridge regression.
    Objective loss of ridge regression is a
    L(X, y, w) = ||Xw - y||^2_2 + 0.5 * alpha * ||w||^2_2
    and exact solution w = X^T @ (X @ X^T + alpha * I)^-1 @ y.
    Here we can make partial kernelization X @ X^T -> K(X, X).

    Parameters
    ----------
    alpha : float (default: 1.0)
        penalty parameter
    kernel : str (default: "linear")
        kernel instead of inner product
    kernel_kwargs : kwargs
        parameters of kernel
    """
    def __init__(self, alpha=1.0, kernel="linear", **kernel_kwargs):
        self.alpha = alpha
        self.kernel = kernels[kernel](**kernel_kwargs)

    def _fit(self, X, y):
        Gram = self.kernel(X, X)
        n_samples = Gram.shape[0]

        ravel = False
        if len(y.shape) == 1:
            y = y[:, np.newaxis]
            ravel = True
        
        Gram.flat[::n_samples + 1] += self.alpha

        try:
            self.dual_coef_ = np.linalg.solve(Gram, y)
        except np.linalg.LinAlgError:
            self.dual_coef_ = np.linalg.lstsq(Gram, y)[0]

        # put back Gram matrix
        Gram.flat[::n_samples + 1] -= self.alpha

        if ravel:
            self.dual_coef_ = self.dual_coef_.ravel()

        self.X_fit_ = X

    def _predict(self, X):
        Gram = self.kernel(X, self.X_fit_)
        return Gram @ self.dual_coef_


import numpy as np
from copy import copy
from mlhandmade.base.base import BaseEstimator

class OVR(BaseEstimator):
    """
    Implementation of One-vs-Rest multiclass strategy.
    This method trains K (number of classes) binary
    classifiers and train all them on full dataset.

    Parameters
    ----------
    classifier : BaseEstimator object
        an estimator with decision_function method
    """
    def __init__(self, classifier):
        if not hasattr(classifier, "decision_function"):
            raise ValueError("This estimator can not be used with OVR.")
        self.classifier = classifier
    
    def _fit(self, X: np.ndarray, y: np.ndarray):
        self.labels = np.unique(y)
        self.num = self.labels.shape[0]
        y_groups = np.array([np.where(y == label, 1, -1) for label in self.labels])
        self.classifiers_ = [copy(self.classifier) for _ in range(self.num)]
        for cl, y_group in zip(self.classifiers_, y_groups):
            cl.fit(X, y_group)
    
    def _predict(self, X: np.ndarray):
        all_outputs = np.array([classifier.decision_function(X) for classifier in self.classifiers_])
        return self.labels.take(np.argmax(all_outputs, axis=0), axis=0)


class OVO(BaseEstimator):
    """
    Implementation of One-vs-One multiclass strategy.
    This method trains K(K-1)/2 (K - number of classes) binary
    classifiers and train all them on pairwise datasets, so 
    these parts smaller than whole dataset and this method can
    be even faster than OVR.

    Parameters
    ----------
    classifier : BaseEstimator object
        an estimator with decision_function method
    """
    def __init__(self, classifier):
        if not hasattr(classifier, "decision_function"):
            raise ValueError("This estimator can not be used with OVO.")
        self.classifier = classifier

    def _fit(self, X: np.ndarray, y: np.ndarray):
        self.labels = np.unique(y)
        self.num = self.labels.shape[0]
        self.classifiers_ = [copy(self.classifier) for _ in range((self.num * (self.num - 1)) // 2)]
        for i in range(self.num):
            for j in range(i+1, self.num):
                X_pair = X[(y == self.labels[i]) | (y == self.labels[j])]
                y_pair = y[(y == self.labels[i]) | (y == self.labels[j])]
                y_pair = np.where(y_pair == self.labels[i], 1, -1)
                self.classifiers_[i + j - 1].fit(X_pair, y_pair)
    
    def _predict(self, X: np.ndarray):
        votes = np.zeros((X.shape[0], self.num))
        confidences = np.zeros((X.shape[0], self.num))
        for i in range(self.num):
            for j in range(i+1, self.num):
                z = self.classifiers_[i + j - 1].predict(X)
                votes[z == 1, i] += 1
                votes[z == -1, j] += 1
                confidences[:, i] += z
                confidences[:, j] -= z
        transformed_confidences = confidences / (3 * (np.abs(confidences) + 1))
        total_votes = np.argmax(votes + transformed_confidences, axis=1)
        return self.labels.take(total_votes, axis=0)


from abc import abstractmethod
import numpy as np

class BaseLoss(object):
    def _set_params(self, X, y, w):
        if not isinstance(X, np.ndarray):
            X = np.array(X)

        if X.size == 0:
            raise ValueError("X is empty")

        if not isinstance(y, np.ndarray):
            y = np.array(y)
        
        if y.size == 0:
            raise ValueError("y is empty")

        if not isinstance(w, np.ndarray):
            w = np.array(w)

        if w.size == 0:
            raise ValueError("w is empty")

        self.X = X
        self.y = y
        self.w = w

    def loss(self, X, y, w):
        self._set_params(X, y, w)
        return self._loss(self.X, self.y, self.w)

    @abstractmethod
    def _loss(self, X, y, w):
        raise NotImplementedError()

    def grad(self, X, y, w):
        self._set_params(X, y, w)
        return self._grad(self.X, self.y, self.w)

    @abstractmethod
    def _grad(self, X, y, w):
        raise NotImplementedError()

class PerceptronLoss(BaseLoss):
    @staticmethod
    def _loss(X, y, w):
        z = y * (X @ w)
        return np.where(-z > 0, -z, 0)

    @staticmethod
    def _grad(X, y, w):
        z = y * (X @ w)
        der = np.where(-z > 0, -1, 0)
        if z.ndim == 0:
            return X * (y * der)
        else:
            return X * (y * der)[:, np.newaxis]

class LogisticLoss(BaseLoss):
    @staticmethod
    def _loss(X, y, w):
        z = y * (X @ w)
        return np.log(1 + np.exp(-z))
    
    @staticmethod
    def _grad(X, y, w):
        z = y * (X @ w)
        der = -np.exp(-z) / (1 + np.exp(-z))
        if z.ndim == 0:
            return X * (y * der)
        else:
            return X * (y * der)[:, np.newaxis]

class HingeLoss(BaseLoss):
    @staticmethod
    def _loss(X, y, w):
        z = y * (X @ w)
        return np.where(1 - z > 0, 1 - z, 0)

    @staticmethod
    def _grad(X, y, w):
        z = y * (X @ w)
        der = np.where(1 - z > 0, -1, 0)
        if z.ndim == 0:
            return X * (y * der)
        else:
            return X * (y * der)[:, np.newaxis]

class SigmoidLoss(BaseLoss):
    @staticmethod
    def _loss(X, y, w):
        z = y * (X @ w)
        return 2 / (1 + np.exp(z))
    
    @staticmethod
    def _grad(X, y, w):
        z = y * (X @ w)
        der = -2 * np.exp(z) / (1 + np.exp(z))**2
        if z.ndim == 0:
            return X * (y * der)
        else:
            return X * (y * der)[:, np.newaxis]

class ExpLoss(BaseLoss):
    @staticmethod
    def _loss(X, y, w):
        z = y * (X @ w)
        return np.exp(-z)

    @staticmethod
    def _grad(X, y, w):
        z = y * (X @ w)
        der = -np.exp(-z)
        if z.ndim == 0:
            return X * (y * der)
        else:
            return X * (y * der)[:, np.newaxis]


from mlhandmade.discriminant_analysis.lda import LinearDiscriminantAnalysis
from mlhandmade.discriminant_analysis.qda import QuadraticDiscriminantAnalysis

__all__ = [
    "LinearDiscriminantAnalysis",
    "QuadraticDiscriminantAnalysis"
]


from abc import abstractmethod
import numpy as np

class BaseEstimator(object):
    """
    Base class for all estimators

    Attributes
    ----------
    y_required : bool (default: True)
        True if y is required for fit, it may be false for unsupervised
        learning algorithms
    fit_required : bool (default: True)
        True if fit required before predict
    """
    y_required = True
    fit_required = True

    def _set_params(self, X, y=None):
        """
        Params validation for estimator

        Parameters
        ----------
        X : array-like of shape (N, D)
            array of N objects with D features
        y : array-like of shape (N,)
            array of target variables
        """
        if not isinstance(X, np.ndarray):
            X = np.asarray(X)
        
        if X.size == 0:
            raise ValueError('X is an empty array.')
        
        if X.ndim == 1:
            self.n_samples, self.n_features = 1, X.shape[0]
        else:
            self.n_samples, self.n_features = X.shape[0], np.prod(X.shape[1:])

        if self.y_required:
            if y is None:
                raise ValueError('y is required for this estimator.')

            if not isinstance(y, np.ndarray):
                y = np.asarray(y)
            
            if y.size == 0:
                raise ValueError('y is an empty array.')

        return X, y
    
    def fit(self, X, y=None, **fit_params):
        X, y = self._set_params(X, y)
        self.fit_required = False
        return self._fit(X, y, **fit_params)

    @abstractmethod
    def _fit(self, X, y, **fit_params):
        raise NotImplementedError()

    def predict(self, X=None):
        if not isinstance(X, np.ndarray):
            X = np.asarray(X)
        
        if X.ndim != 2:
            X = np.atleast_2d(X)

        if X is not None and not self.fit_required:
            return self._predict(X)
        else:
            raise ValueError("Fit is necessary for this estimator.")
    
    @abstractmethod
    def _predict(self, X=None):
        raise NotImplementedError()

    def transform(self, X=None):
        if not isinstance(X, np.ndarray):
            X = np.asarray(X)
        
        if X.ndim != 2:
            X = np.atleast_2d(X)

        if X is not None and not self.fit_required:
            return self._transform(X)
        else:
            raise ValueError("Fit is necessary for this estimator.")

    @abstractmethod
    def _transform(self, X=None):
        raise NotImplementedError()


from mlhandmade.plotting.confusion_matrix import plot_confusion_matrix
from mlhandmade.plotting.scatter import scatter_plot
from mlhandmade.plotting.decision_regions import plot_decision_regions

__all__ = [
    "plot_confusion_matrix",
    "scatter_plot",
    "plot_decision_regions"
]


import numpy as np

from ..base import BaseEstimator
from .losses import *
from .optimizers import *
from ..preprocessing import add_bias_feature, onehot
from ..utils.validations import check_random_state

loss_dict = {"perceptron" : PerceptronLoss, "logistic" : LogisticLoss,
             "hinge" : HingeLoss, "sigmoid" : SigmoidLoss}

optimizer_dict = {"gd" : GD, "sgd" : SGD, "batch_gd" : BatchGD, "sag" : SAG}

class LinearClassifier(BaseEstimator):
    """
    Linear classifier class

    Attributes
    ----------
    loss : str
        Loss function to minimize.
        Availible losses:
        ["perceptron", "logistic", "hinge", "sigmoid"]
    optimizer : str
        Optimization algorithm to minimize loss
        Availible optimizers:
        ["gd", "sgd", "batch_gd", "sag"]
    epochs : int
        number of optimization iterations
    tol : float
        numerical tolerance
    random_state : int
        random state to debug calculations
    optimizer_settings : kwargs
        kwargs for chosen optimizer
    """
    def __init__(self, loss, optimizer, epochs, tol=1e-3, random_state=0, **optimizer_settings):
        self.loss = loss_dict[loss]()
        self.optimizer = optimizer_dict[optimizer](**optimizer_settings)
        self.epochs = epochs
        self.tol = tol
        self.rgen = check_random_state(random_state)

    def _fit(self, X: np.ndarray, y: np.ndarray):
        X = add_bias_feature(X)
        self.w_ = self.rgen.normal(loc=0.0, scale=0.01, size=self.n_features + 1)
        for _ in range(self.epochs):
            self.w_ = self.optimizer.update(self.loss.grad, X, y, self.w_)
            if np.mean(self.loss._loss(X, y, self.w_)) < self.tol:
                break

    def decision_function(self, X: np.ndarray):
        return X @ self.w_[1:] + self.w_[0]

    def _predict(self, X):
        return np.where(self.decision_function(X) > 0, 1, -1)

class SoftmaxRegressor(BaseEstimator):
    """
    Softmax regression class (multinominal logistic regression).
    Softmax function can't be represented as finite sum,
    so we can't use any stochastic descents here.

    Attributes
    ----------
    eta : float
        learning rate
    epochs : int
        number of optimization iterations
    random_state : int
        random state to debug calculations
    """
    def __init__(self, eta, epochs, random_state=0):
        self.eta = eta
        self.epochs = epochs
        self.rgen = check_random_state(random_state)

    @staticmethod
    def _softmax(X, w):
        a = X @ w
        exp_a = np.exp(a - np.max(a, axis=-1, keepdims=True))
        return exp_a / np.sum(exp_a, axis=-1, keepdims=True)

    def _fit(self, X, y):
        X = add_bias_feature(X)
        if y.ndim == 1:
            y = onehot(y)
        self.w_ = self.rgen.normal(loc=0.0, scale=0.01, size=(self.n_features + 1, y.shape[1]))
        for _ in range(self.epochs):
            grad = X.T @ (self._softmax(X, self.w_) - y)
            self.w_ -= self.eta * grad

    def _predict(self, X: np.ndarray):
        X = add_bias_feature(X)
        return np.argmax(self._softmax(X, self.w_), axis=-1)


import numpy as np

from ..base import BaseEstimator

def _class_means(X, y):
    classes, y = np.unique(y, return_inverse=True)
    cnt = np.bincount(y)
    means = np.zeros(shape=(len(classes), X.shape[1]))
    np.add.at(means, y, X)
    means /= cnt[:, np.newaxis]
    return means

class LinearDiscriminantAnalysis(BaseEstimator):
    def __init__(self, tol=1e-4, n_components=None):
        self.tol = tol
        self.n_components = n_components

    def _solve_svd(self, X, y):
        n_samples, _ = X.shape
        n_classes = len(self.classes_)

        self.means_ = _class_means(X, y)

        Xc = []
        for idx, group in enumerate(self.classes_):
            Xg = X[y == group, :]
            Xc.append(Xg - self.means_[idx])
        
        self.xbar_ = np.dot(self.priors_, self.means_)

        Xc = np.concatenate(Xc, axis=0) # ravel in 2d case

        std = Xc.std(axis=0)
        std[std == 0] = 1.0 # avoid zero division
        fac = 1.0 / (n_samples - n_classes)

        # scaling of data
        X = np.sqrt(fac) * (Xc / std)

        _, S, Vt = np.linalg.svd(X, full_matrices=False)
        rank = np.sum(S > self.tol)
        scalings = (Vt[:rank] / std).T / S[:rank]
        fac = 1.0 if n_classes == 1 else 1.0 / (n_classes - 1)

        X = np.dot(
            (
                (np.sqrt((n_samples * self.priors_) * fac))
                * (self.means_ - self.xbar_).T
            ).T,
            scalings
        )
        
        _, S, Vt = np.linalg.svd(X, full_matrices=False)

        rank = np.sum(S > self.tol * S[0])
        self.scalings_ = np.dot(scalings, Vt.T[:, :rank])
        coef = np.dot(self.means_ - self.xbar_, self.scalings_)
        self.intercept_ = -0.5 * np.sum(coef**2, axis=1) + np.log(self.priors_)
        self.coef_ = np.dot(coef, self.scalings_.T)
        self.intercept_ -= np.dot(self.xbar_, self.coef_.T)

    def _fit(self, X, y):
        self.classes_ = np.unique(y)

        # prior class probabilities from sample data
        _, y_t = np.unique(y, return_inverse=True)
        self.priors_ = np.bincount(y_t) / float(len(y))

        # maximum possible components to reduce
        max_components = min(len(self.classes_) - 1, X.shape[1])

        if self.n_components is None:
            self._max_components = max_components
        else:
            if self.n_components > max_components:
                raise ValueError(
                    "n_components cannot be larger than min(n_features, n_classes - 1)."
                )
            self._max_components = self.n_components

        self._solve_svd(X, y)

        if self.classes_.size == 2:
            self.coef_ = np.array(
                self.coef_[1, :] - self.coef_[0, :], ndmin=2, dtype=X.dtype
            )
            self.intercept_ = np.array(
                self.intercept_[1] - self.intercept_[0], ndmin=1, dtype=X.dtype
            )

        self.n_features_out = self._max_components

    def _transform(self, X):
        X_new = np.dot(X - self.xbar_, self.scalings_)
        return X_new[:, :self._max_components]

    @staticmethod
    def _softmax(decisions):
        exp_a = np.exp(decisions - np.max(decisions, axis=-1, keepdims=True))
        return exp_a / np.sum(exp_a, axis=-1, keepdims=True)

    def decision_function(self, X):
        scores = X @ self.coef_.T + self.intercept_
        return scores.ravel() if scores.shape[1] == 1 else scores

    def predict_proba(self, X):
        decisions = self.decision_function(X)
        if self.classes_.size == 2:
            proba = 1.0 / (1 + np.exp(-decisions))
            return np.vstack([1 - proba, proba]).T
        else:
            return self._softmax(decisions)

    def _predict(self, X):
        scores = self.decision_function(X)
        if len(scores.shape) == 1:
            indices = np.where(scores > 0, 1, 0)
        else:
            indices = np.argmax(scores, axis=-1)
        return self.classes_[indices]


import numpy as np
from .kernel import Kernel

class PolynomialKernel(Kernel):
    """
    Polynomial kernel
    k(x, y) = (x @ y + c)^M

    Parameters
    ----------
    const : float
        a constant to be added
    degree : int
        degree of polynomial order
    """
    def __init__(self, degree=2, const=0.0):
        self.const = const
        self.degree = degree

    def __call__(self, x, y, pairwise=True):
        """
        calculate pairwise polynomial kernel

        Parameters
        ----------
        x : (..., ndim) ndarray
            input
        y : (..., ndim) ndarray
            another input with the same shape
            
        Returns
        -------
        output : ndarray
            polynomial kernel
        """
        if pairwise:
            x, y = self._pairwise(x, y)
        return (np.sum(x * y, axis=-1) + self.const) ** self.degree


import numpy as np

from ..base import BaseEstimator
from ..preprocessing import add_bias_feature
from .optimizers import *
from ..utils.validations import check_random_state

optimizer_dict = {"gd" : GD, "sgd" : SGD, "batch_gd" : BatchGD, "sag" : SAG}
exact_methods = ["direct", "svd", "qr"]

class LinearRegressor(BaseEstimator):
    """
    Logistic Regression class implemented with different
    methods: exact(matrix decompositions) and approximate(gradient).
    Exact methods faster due to numpy implementation

    Objective loss:
    L(X, y, w) = ||Xw - y||^2_2

    Attributes
    ----------
    method : str
        Optimization algorithm to minimize loss
        Availible optimizers:
        ["direct", "svd", "qr", "gd", "sgd", "batch_gd", "sag"]
    epochs : int
        number of optimization iterations
    tol : float
        numerical tolerance
    random_state : int
        random state to debug calculations
    optimizer_settings : kwargs
        kwargs for chosen optimizer
    """
    def __init__(
        self,
        method: str,
        epochs: int = None,
        tol: float = 1e-3,
        random_state: int = 0,
        **optimizer_settings
    ) -> None:
        self._validate_input(method, epochs)
        if method in optimizer_dict:
            self.optimizer = optimizer_dict[method](**optimizer_settings)
        self.method = method
        
        self.epochs = epochs
        self.tol = tol
        self.rgen = check_random_state(random_state)

    @staticmethod
    def _validate_input(method, epochs):
        if method in optimizer_dict:
            if epochs is None:
                raise ValueError("epochs must be positive int if method is gradient")
        elif method in exact_methods:
            pass
        else:
            raise ValueError(f"Unknown method, got {method}.\nAvailible methods:\
            \nexact - {exact_methods}\ngradient - {list(optimizer_dict.keys())}")

    @staticmethod
    def _loss_grad(X, y, w):
        if y.ndim == 0:
            return 2 * X * (X @ w - y)
        else:
            return 2 * X.T @ (X @ w - y)

    def _fit(self, X, y):
        X = add_bias_feature(X)
        
        if self.method in optimizer_dict:
            self.w_ = self.rgen.normal(loc=0.0, scale=0.01, size=self.n_features + 1)
            for _ in range(self.epochs):
                self.w_ = self.optimizer.update(self._loss_grad, X, y, self.w_)
                if np.linalg.norm(X @ self.w_ - y) < self.tol:
                    break
        
        elif self.method == "direct":
            self.w_ = np.linalg.inv(X.T @ X) @ X.T @ y

        elif self.method == "svd":
            # w = V x S^-1 x U
            self.w_ = np.linalg.pinv(X) @ y
        
        elif self.method == "qr":
            Q, R = np.linalg.qr(X)
            self.w_ = np.linalg.inv(R) @ Q.T @ y

    def _predict(self, X):
        return X @ self.w_[1:] + self.w_[0]

class RidgeRegressor(BaseEstimator):
    """
    Regularized linear regression implemented with svd

    Objective loss:
    L(X, y, w) = ||Xw - y||^2_2 + alpha * ||w||^2_2

    Attributes
    ----------
    alpha : float
        penalty parameter
    """
    def __init__(
        self,
        alpha: float = 0,
    ) -> None:
        self.alpha = alpha

    @staticmethod
    def _solve_svd(X, y, alpha):
        # w = V x (S^T x S + alpha * I)^-1 x S^T x U^T x y
        U, s, Vt = np.linalg.svd(X, full_matrices=False)
        idx = s > 1e-15 # avoid numerical overflow in denominator
        s_nnz = s[idx]
        UTy = U.T @ y
        d = np.zeros(s.size, dtype=X.dtype)
        d[idx] = s_nnz / (s_nnz**2 + alpha)
        d_UT_y = d * UTy
        return Vt.T @ d_UT_y

    def _fit(self, X, y):
        X = add_bias_feature(X)
        self.w_ = self._solve_svd(X, y, self.alpha)
        
    def _predict(self, X):
        return X @ self.w_[1:] + self.w_[0]


import numpy as np
import matplotlib.pyplot as plt

def scatter_plot(
    X: np.ndarray,
    y: np.ndarray,
    *,
    ax = None,
    legend = 0,
    markers = ('s', 'x', 'o', '^', '*'),
    colors = ('IndianRed', 'RoyalBlue', 'lightgreen', 'grey', 'cyan')
):
    if ax is None:
        ax = plt.gca()

    for idx, cl in enumerate(np.unique(y)):
        ax.scatter( x=X[y == cl, 0],
                    y=X[y == cl, 1],
                    s=40,
                    alpha=0.8,
                    c=colors[idx],
                    marker=markers[idx],
                    label=cl)

    if legend:
        handles, labels = ax.get_legend_handles_labels()
        ax.legend(handles, labels, framealpha=0.3, scatterpoints=1, loc=legend)

    return ax


import numpy as np

def get_split_mask(X, feature, value):
    left_mask = X[:, feature] < value
    right_mask = X[:, feature] >= value
    return left_mask, right_mask

def find_splits(X):
    X_uniq = np.unique(X)
    return (X_uniq[1:] + X_uniq[:-1]) / 2

def get_best_split(criterion, X, y, features, sample_weight):
    min_gain = np.inf
    best_feature = None
    best_value = None

    for feature in features:
        split_values = find_splits(X[:, feature])
        for value in split_values:
            left_mask, right_mask = get_split_mask(X, feature, value)
            splits = (y[left_mask], y[right_mask])
            splits_weight = (sample_weight[left_mask], sample_weight[right_mask])

            if not np.any(splits_weight[0]) or not np.any(splits_weight[1]):
                continue

            gain = criterion.gain(y, splits, splits_weight)

            if gain < min_gain:
                min_gain = gain
                best_feature = feature
                best_value = value

    return min_gain, best_feature, best_value


import numpy as np
from itertools import combinations_with_replacement as combinations_w_r
from itertools import combinations

def data_shuffle(X: np.ndarray, y: np.ndarray = None, rgen=None) -> np.ndarray:
    idx = np.arange(X.shape[0])
    if rgen is None:
        np.random.shuffle(idx)
    else:
        rgen.shuffle(idx)
    if y is None:
        return X[idx]
    else:
        return X[idx], y[idx]

def ordinal(y: np.ndarray) -> np.ndarray:
    labels = np.unique(y)
    ordinal = np.zeros(y.shape, dtype=np.int64)
    for idx, cl in enumerate(labels):
        ordinal[y == cl] = idx
    return ordinal

def binary(y: np.ndarray, pos_label: int = 1, neg_label: int = -1) -> np.ndarray:
    labels = np.unique(y)
    if labels.shape[0] != 2:
        raise ValueError("Must be two classes to encode")
    else:
        return np.where(y == labels[0], pos_label, neg_label).astype(int)

def add_bias_feature(X: np.ndarray) -> np.ndarray:
    if X.ndim == 1:
        return np.insert(X, 0, 1)
    else:
        # np.hstack((np.ones((X.shape[0], 1)), X))
        return np.pad(X, [(0, 0), (1, 0)], mode="constant", constant_values=1)

def standardize(X: np.ndarray) -> np.ndarray:
    return (X - np.mean(X, axis=0)) / np.std(X, axis=0)

def onehot(y: np.ndarray) -> np.ndarray:
    n_values = np.max(y) + 1
    return np.eye(n_values, dtype=np.int64)[y]

def polynomial_features(X: np.ndarray, degree: int, interactions_only=False):
    n_samples, n_features = X.shape
    combinator = combinations if interactions_only else combinations_w_r
    combs = [combinator(range(n_features), i) for i in range(1, degree + 1)]
    product_tuples = [item for sublist in combs for item in sublist]
    n_output_features = len(product_tuples)

    XP = np.empty((n_samples, n_output_features))
    for i, idx_product in enumerate(product_tuples):
        XP[:, i] = np.prod(X[:, idx_product], axis=1)
    
    return XP


from abc import abstractmethod
import numpy as np

from ..base import BaseEstimator
from .criterion import ClassificationCriterion, RegressionCriterion
from .splitter import get_best_split, get_split_mask
from ..utils.validations import check_random_state, check_sample_weight

class BaseDecisionTree(BaseEstimator):
    """
    Implementation of Decision tree
    with greedy algorithm

    Parameters
    ----------
    is_classifier : bool
        True if it is a classification tree
    criterion : BaseCriterion object
        callable criterion obeject
    max_feauters : int (default: None)
        number of features on each recursion step
    min_samples_leaf : int (default: 1)
        minimum possible samples in each leaf
    max_depth : int (default: inf)
        maximum depth of tree
    n_features : int (default: None)
        number of feauters, need for recursion
    n_classes : int (default: None)
        number of classes, need for recursion
    """
    def __init__(
        self,
        is_classifier,
        criterion,
        max_features=None,
        min_samples_leaf=1,
        max_depth=np.inf,
        random_state=0,
        n_features=None,
        n_classes=None
    ):
        self.is_classifier = is_classifier
        self.criterion = criterion
        self.max_features = max_features
        self.min_samples_leaf = min_samples_leaf
        self.max_depth = max_depth
        self.rgen = check_random_state(random_state)

        self.n_features = n_features
        self.n_classes = n_classes
        self.is_leaf = False

    def _fit(self, X, y, sample_weight=None):
        """
        Fit decision tree.

        Parameters
        ----------
        X : np.ndarray of shape (N, D)
            array of data
        y : np.ndarray of shape (N,)
            array of targets
        sample_weight : array-like of shape (N,), number, None
            array of weights for all statistics: average, probability, median
            sample_weight is needed for random forest bootstrap implementation
        """
        if self.is_classifier:
            self.classes_ = np.unique(y)
            self.n_classes = self.classes_.shape[0]
        sample_weight = check_sample_weight(sample_weight, X)
        self._build_tree(X, y, sample_weight)

    def _build_tree(self, X, y, sample_weight):
        try:
            assert X.shape[0] > self.min_samples_leaf
            assert self.max_depth > 0

            if self.max_features is None:
                self.max_features = self.n_features
            elif self.max_features == "sqrt":
                self.max_features = round(np.sqrt(self.n_features))
            
            features = self.rgen.choice(self.n_features, self.max_features)
            gain, feature, value = get_best_split(self.criterion, X, y, features, sample_weight)
            assert gain != np.inf
            
            self.feature = feature
            self.value = value

            left_mask, right_mask = get_split_mask(X, feature, value)

            self.left = BaseDecisionTree(
                self.is_classifier,
                self.criterion,
                self.max_features,
                self.min_samples_leaf,
                self.max_depth - 1,
                self.rgen,
                self.n_features,
                self.n_classes
            )
            self.left._build_tree(X[left_mask], y[left_mask], sample_weight[left_mask])

            self.right = BaseDecisionTree(
                self.is_classifier,
                self.criterion,
                self.max_features,
                self.min_samples_leaf,
                self.max_depth - 1,
                self.rgen,
                self.n_features,
                self.n_classes
            )
            self.right._build_tree(X[right_mask], y[right_mask], sample_weight[right_mask])
        except AssertionError:
            self.is_leaf = True
            self.calculate_leaf_value(y, sample_weight)

    def calculate_leaf_value(self, y, sample_weight):
        if self.is_classifier:
            self.leaf_value = np.bincount(y, weights=sample_weight, minlength=self.n_classes) / np.sum(sample_weight)
        else:
            self.leaf_value = np.average(y, weights=sample_weight)

    def predict_sample(self, x):
        if not self.is_leaf:
            if x[self.feature] < self.value:
                return self.left.predict_sample(x)
            else:
                return self.right.predict_sample(x)
        else:
            return self.leaf_value

    @abstractmethod
    def _predict(self, X):
        raise NotImplementedError()

class DecisionTreeClassifier(BaseDecisionTree):
    """
    Decision tree classifier class.

    Parameters
    ----------
    criterion : ["gini", "entropy"] (default: "gini")
        impurity criterion
    max_features : int (default: None)
        number of features on each recursion step
    min_samples_leaf : int (default: 1)
        minimum possible samples in each leaf
    max_depth : int (default: inf)
        maximum depth of tree
    """
    def __init__(
        self,
        *,
        criterion="gini",
        max_features=None,
        min_samples_leaf=1,
        max_depth=np.inf,
        random_state=0
    ):
        is_classifier = True
        criterion = ClassificationCriterion(criterion)
        super().__init__(
            is_classifier,
            criterion,
            max_features,
            min_samples_leaf,
            max_depth,
            random_state
        )

    def predict_proba(self, X):
        probas = np.zeros((X.shape[0], self.n_classes))
        for i, x in enumerate(X):
            probas[i] = self.predict_sample(x)
        return probas

    def _predict(self, X):
        return self.classes_.take(np.argmax(self.predict_proba(X), axis=-1), axis=0)

class DecisionTreeRegressor(BaseDecisionTree):
    """
    Decision tree regressor class.

    Parameters
    ----------
    criterion : ["mse", "mae"] (default: "mse")
        impurity criterion
    max_features : int (default: None)
        number of features on each recursion step
    min_samples_leaf : int (default: 1)
        minimum possible samples in each leaf
    max_depth : int (default: inf)
        maximum depth of tree
    """
    def __init__(
        self,
        *,
        criterion="mse",
        max_features=None,
        min_samples_leaf=1,
        max_depth=np.inf,
        random_state=0
    ):
        is_classifier = False
        criterion = RegressionCriterion(criterion)
        super().__init__(
            is_classifier,
            criterion,
            max_features,
            min_samples_leaf,
            max_depth,
            random_state
        )

    def _predict(self, X):
        y_pred = np.zeros(X.shape[0])
        for i, x in enumerate(X):
            y_pred[i] = self.predict_sample(x)
        return y_pred


import numpy as np

class Kernel(object):
    """
    Base class for kernel function
    """

    def _pairwise(self, x, y):
        """
        all pairs of x and y
        Parameters
        ----------
        x : (sample_size, n_features)
            input
        y : (sample_size, n_features)
            another input
        Returns
        -------
        output : tuple
            two array with shape (sample_size, sample_size, n_features)
        """
        return (
            np.tile(x, (len(y), 1, 1)).transpose(1, 0, 2),
            np.tile(y, (len(x), 1, 1))
        )


import numpy as np

from ..base import BaseEstimator

class QuadraticDiscriminantAnalysis(BaseEstimator):
    def __init__(self, tol=1e-4, reg_param=0.0):
        self.tol = tol
        self.reg_param = reg_param

    def _fit(self, X, y):
        n_samples, _ = X.shape
        self.classes_, y = np.unique(y, return_inverse=True)
        n_classes = len(self.classes_)

        self.priors_ = np.bincount(y) / float(n_samples)
        means = []
        scalings = []
        rotations = []
        for ind in range(n_classes):
            Xg = X[y == ind, :]
            meang = np.mean(Xg, axis=0)
            means.append(meang)
            Xgc = Xg - meang
            _, S, Vt = np.linalg.svd(Xgc, full_matrices=False)

            S2 = S**2 / (len(Xg) - 1)
            S2 = ((1 - self.reg_param) * S2) + self.reg_param

            scalings.append(S2)
            rotations.append(Vt.T)
        self.means_ = np.asarray(means)
        self.scalings_ = np.asarray(scalings)
        self.rotations_ = np.asarray(rotations)

    def _decision_function(self, X):
        norm2 = []
        for i in range(len(self.classes_)):
            R = self.rotations_[i]
            S = self.scalings_[i]
            Xm = X - self.means_[i]
            X2 = np.dot(Xm, R * (S ** (-0.5)))
            norm2.append(np.sum(X2**2, axis=1))
        norm2 = np.asarray(norm2).T
        u = np.asarray([np.sum(np.log(s)) for s in self.scalings_])
        return -0.5 * (norm2 + u) + np.log(self.priors_)

    def decision_function(self, X):
        decisions = self._decision_function(X)
        if len(self.classes_) == 2:
            return decisions[:, 1] - decisions[:, 0]
        return decisions

    @staticmethod
    def _softmax(values):
        exp_a = np.exp(values - np.max(values, axis=-1, keepdims=True))
        return exp_a / np.sum(exp_a, axis=-1, keepdims=True)

    def _predict(self, X=None):
        decisions = self._decision_function(X)
        y_pred = self.classes_.take(np.argmax(decisions, axis=1))
        return y_pred

    def predict_proba(self, X):
        values = self._decision_function(X)
        return self._softmax(values)


import numpy as np
from numpy.typing import ArrayLike

from ..preprocessing import data_shuffle
from ..utils.validations import check_random_state

class KFoldCV:
    """
    k-fold cross-validation.
    This class allows to iterate over
    train and test indices for each cross
    validation fold

    Parameters
    ----------
    n_splits : int (default: 5)
        number of splits in each fold
    shuffle : bool (default: False)
        shuffle data before iterations
    random_state : int (default: 0)
        seed for debugging
    """
    def __init__(
        self,
        n_splits: int = 5,
        shuffle: bool = False,
        random_state: int = 0
    ):
        self.n_splits = n_splits
        self.shuffle = shuffle
        self.rgen = check_random_state(random_state)

    def _validate_params(self, X, y):
        if not isinstance(X, np.ndarray):
            X = np.asarray(X)
        
        if X.ndim < 2:
            X = np.atleast_2d(X)

        if y is not None:
            if not isinstance(y, np.ndarray):
                y = np.asarray(y)
            
            if y.ndim < 2:
                y = np.atleast_2d(y)

    def split(self, X: np.ndarray, y: np.ndarray = None):
        """
        Get iterator over all folds

        Parametrs
        ---------
        X : ndarray of shape (N, D)
            ndarray of input samples
        y : ndarray of shape (N,) (default: None)
            ndarray of input targets
        
        Yields
        ------
        train_index, test_index : ndarrays
            indices for train and test parts of folds
        """
        self._validate_params(X, y)
        self.num_samples_ = X.shape[0]
        indices = np.arange(self.num_samples_)
        for test_index in self._iter_test_masks(X, y):
            train_index = indices[np.logical_not(test_index)]
            test_index = indices[test_index]
            yield train_index, test_index

    def _iter_test_masks(self, X: np.ndarray, y: np.ndarray):
        for test_index in self._iter_test_indices(X, y):
            test_mask = np.zeros(self.num_samples_, dtype=bool)
            test_mask[test_index] = True
            yield test_mask

    def _iter_test_indices(self, X: np.ndarray, y: np.ndarray = None):
        indices = np.arange(self.num_samples_)
        if self.shuffle:
            indices = data_shuffle(indices, rgen=self.rgen)
        
        fold_sizes = np.full(self.n_splits, self.num_samples_ // self.n_splits, dtype=int)
        fold_sizes[:self.num_samples_ % self.n_splits] += 1
        current = 0
        for fold_size in fold_sizes:
            start, stop = current, current + fold_size
            yield indices[start:stop]
            current = stop

def cross_val_score(
    estimator,
    X,
    y,
    *,
    score=None,
    cv=5,
    shuffle=False,
    random_state=0,
    **score_kwargs
):
    if score is None:
        raise ValueError("Choose availible score.")

    scores = np.zeros(cv)
    kf = KFoldCV(n_splits=cv, shuffle=shuffle, random_state=random_state)
    for score_idx, (train_idx, test_idx) in enumerate(kf.split(X)):
        estimator.fit(X[train_idx], y[train_idx])
        y_pred = estimator.predict(X[test_idx])
        scores[score_idx] = score(y[test_idx], y_pred, **score_kwargs)
    return np.mean(scores)


import numpy as np
import matplotlib.pyplot as plt
from matplotlib.colors import ListedColormap

def plot_decision_regions(
    X: np.ndarray,
    y: np.ndarray,
    classifier,
    *,
    ax = None,
    legend = 0,
    resolution = 0.02,
    margin = 0.25,
    markers = ('s', 'x', 'o', '^', '*'),
    colors = ('IndianRed', 'RoyalBlue', 'lightgreen', 'gray', 'cyan')
):
    if ax is None:
        ax = plt.gca()

    cmap = ListedColormap(colors[:len(np.unique(y))])
    
    x1_min, x1_max = X[:, 0].min() - margin, X[:, 0].max() + margin
    x2_min, x2_max = X[:, 1].min() - margin, X[:, 1].max() + margin
    
    xx1, xx2 = np.meshgrid(np.arange(x1_min, x1_max, resolution), np.arange(x2_min, x2_max, resolution))
    
    Z = classifier.predict(np.array([xx1.ravel(), xx2.ravel()]).T)
    Z = Z.reshape(xx1.shape)
    
    ax.contourf(xx1, xx2, Z, alpha=0.3, cmap=cmap)
    ax.set_xlim(x1_min, x1_max)
    ax.set_ylim(x2_min, x2_max)
    
    for idx, cl in enumerate(np.unique(y)):
        ax.scatter( x=X[y == cl, 0],
                    y=X[y == cl, 1],
                    s=40,
                    alpha=0.8,
                    c=colors[idx],
                    marker=markers[idx],
                    label=cl)

    if legend:
        handles, labels = ax.get_legend_handles_labels()
        ax.legend(handles, labels, framealpha=0.3, scatterpoints=1, loc=legend)

    return ax


import numpy as np
from itertools import product

from ..utils.validations import check_consistent_length

def accuracy_score(y_true, y_pred, *, normalize=True, sample_weight=None):
    if not isinstance(y_true, np.ndarray):
        y_true = np.asarray(y_true)

    if not isinstance(y_pred, np.ndarray):
        y_pred = np.asarray(y_pred)

    if y_true.size != y_pred.size:
        raise ValueError("y_true and y_pred must be equal size.")
    
    check_consistent_length(y_true, y_pred, sample_weight)

    score = y_true == y_pred

    if normalize:
        return np.average(score, weights=sample_weight)
    elif sample_weight is not None:
        return np.dot(score, sample_weight)
    else:
        return np.sum(score)

def confusion_matrix(y_true, y_pred, *, labels=None,  normalize=None, sample_weight=None):
    if not isinstance(y_true, np.ndarray):
        y_true = np.asarray(y_true)

    if not isinstance(y_pred, np.ndarray):
        y_pred = np.asarray(y_pred)
    
    if y_true.size != y_pred.size:
        raise ValueError("y_true and y_pred must be equal size.")

    if labels is None:
        labels = np.unique(y_true)
    else:
        labels = np.asarray(labels)
        n_labels = labels.size
        if n_labels == 0:
            raise ValueError("`labels` must contain at least one label")
        elif y_true.size == 0:
            return np.zeros((n_labels, n_labels), dtype=int)
        elif len(np.intersect1d(y_true, labels)) == 0:
            raise ValueError("At least one label specified in `labels` must be in `y_true`")
    
    if sample_weight is None:
        sample_weight = np.ones(y_true.shape[0], dtype=np.int64)
    else:
        sample_weight = np.asarray(sample_weight)

    check_consistent_length(y_true, y_pred, sample_weight)

    if normalize not in ["true", "pred", "all", None]:
        raise ValueError("normalize must be one of {'true', 'pred', 'all', None}")

    n_labels = labels.size
    # `y_true` and `y_pred` should be indeces,
    # i.e 0, 1, 2, ..., so check if conversion
    # is needed
    need_index_conversion = not(
        labels.dtype.kind in {"i", "u", "b"} # dtype of `labels` must be int, uint or bool
        and np.all(labels == np.arange(n_labels)) # `labels` equals [0, 1, ..., n_labels - 1]
        and y_true.min() >= 0 # all elements of `y_true` are >= 0
        and y_pred.min() >= 0 # all elements of `y_pred` are >= 0
    )
    if need_index_conversion:
        # dict {label : label ordinal index}
        label_to_ind = {y: x for x, y in enumerate(labels)}
        # convert `y_pred` and `y_true` to indices corresponding to `labels`
        # if element not found in labels than index will be `n_labels + 1`
        y_true = np.array([label_to_ind.get(x, n_labels + 1) for x in y_true])
        y_pred = np.array([label_to_ind.get(x, n_labels + 1) for x in y_pred])

    # intersect y_pred, y_true with labels, eliminate items not in labels
    ind = np.logical_and(y_true < n_labels, y_pred < n_labels)
    if not np.all(ind):
        y_true = y_true[ind]
        y_pred = y_pred[ind]
        # `sample_weight` must be the same size as `y_true` and `y_pred`
        sample_weight = sample_weight[ind]

    if sample_weight.dtype.kind in {"i", "u", "b"}:
        dtype = np.int64
    else:
        dtype = np.float64

    # from scipy.sparse import coo_matrix
    # cm = coo_matrix(
    #     (sample_weight, (y_true, y_pred)),
    #     shape=(n_labels, n_labels),
    #     dtype=dtype,
    # ).toarray()

    cm = np.zeros(shape=(n_labels, n_labels), dtype=dtype)
    for w, t_true, t_pred in zip(sample_weight, y_true, y_pred):
        cm[t_true, t_pred] += w

    # context manager to avoid "same_kind" rule cast error
    # don't use `/=` operator for the same reason
    with np.errstate(all="ignore"):
        if normalize == "true":
            cm = cm / cm.sum(axis=1, keepdims=True)
        elif normalize == "pred":
            cm = cm / cm.sum(axis=0, keepdims=True)
        elif normalize == "all":
            cm = cm / cm.sum()

    return cm


import numpy as np
import heapq
import itertools

class KDTreeBase:
    """
    kd-tree implementation class.

    Attributes
    ----------
    dim : ndarray of shape(D,)
        dimension of space
    division_axis : int
        splitting hyperplane orthoghonal to division axis
    data : ndarray of shape(D,)
        point in current leaf
    values : ndarray of shape(N,)
        values of some function in points, 
        for instance binary classes or continuos value in case of regression
    left, right : KDTreeBase object
        references to child leafs
        left and right doesn't make any sense, just analogy with binary tree
    """
    def __init__(self, points=None, values=None, division_axis=0):
        self.dim = points.shape[-1]
        self.division_axis = division_axis
        
        if points.shape[0] > 1:
            idx_sort = np.argsort(points[:, division_axis])
            points = points[idx_sort]
            values = values[idx_sort]

            # equivalent to `points.shape[0] // 2`
            middle = points.shape[0] >> 1
            division_axis = (division_axis + 1) % self.dim

            self.data = points[middle]
            self.value = values[middle]
            self.left = KDTreeBase(points[:middle], values[:middle], division_axis)
            self.right = KDTreeBase(points[middle + 1:], values[middle + 1:], division_axis)
        elif points.shape[0] == 1:
            self.data = points[0]
            self.value = values[0]
            self.left = None
            self.right = None
        else:
            self.data = None
            self.value = None

    def _query_knn(self, point, k, metric, heap, counter=itertools.count()):
        """
        k nearest neighbor query

        Parameters
        ----------
        point : ndarray of shape(D,)
            get neighbors of this point
        k : int
            number of neighbors
        metric : {1, 2, inf}
            metric of distance for numpy.linalg.norm function.
        heap : list
            heap of neighbors candidates,
            initial value suppose to be []
        counter : itertools.count object
            counter to avoid comparison error in heappushpop()
        """
        if self.data is not None:
            dist = self.distance(self.data, point, metric)
            dx = self.data[self.division_axis] - point[self.division_axis]
            # less distance must have higher priority,
            # so `-dist` in `item` instead of `dist`
            item = (-dist, next(counter), self.data, self.value)
            if len(heap) < k:
                heapq.heappush(heap, item)
            elif dist < -heap[0][0]:
                heapq.heappushpop(heap, item)

            if dx < 0:
                if self.right is not None:
                    self.right._query_knn(point, k, metric, heap, counter)
            else:
                if self.left is not None:
                    self.left._query_knn(point, k, metric, heap, counter)
            if np.abs(dx) < -heap[0][0]:
                if dx >= 0:
                    if self.right is not None:
                        self.right._query_knn(point, k, metric, heap, counter)
                else:
                    if self.left is not None:
                        self.left._query_knn(point, k, metric, heap, counter)

    @staticmethod
    def distance(point1, point2, metric):
        return np.linalg.norm(point1 - point2, ord=metric)

    @property
    def get_dim(self):
        return self.dim


from mlhandmade.model_selection.class_metrics import (
    accuracy_score,
    confusion_matrix
)

from mlhandmade.model_selection.kflod_cv import (
    KFoldCV,
    cross_val_score
)

from mlhandmade.model_selection.grid_search import (
    GridSearchCV
)

__all__ = [
    "accuracy_score",
    "confusion_matrix",
    "KFoldCV",
    "cross_val_score",
    "GridSearchCV"
]


from abc import abstractmethod
import numpy as np
from ..preprocessing import data_shuffle

class BaseOptimizer:
    """
    Base class for optimizers of finite sums.
    """
    @abstractmethod
    def update(self, grad, X, y, w):
        raise NotImplementedError()

class GD(BaseOptimizer):
    def __init__(self, eta):
        self.eta = eta

    def update(self, grad, X, y, w):
        w -= self.eta * np.mean(grad(X, y, w), axis=0)
        return w

class SGD(BaseOptimizer):
    def __init__(self, eta):
        self.eta = eta

    def update(self, grad, X, y, w):
        X, y = data_shuffle(X, y)
        for xi, target in zip(X, y):
            w -= self.eta * grad(xi, target, w)
        return w

class BatchGD(BaseOptimizer):
    def __init__(self, eta, batch_size, random_state=1):
        self.eta = eta
        self.batch_size = batch_size
        self.rgen = self._setup_rgen(random_state)

    @staticmethod
    def _setup_rgen(random_state):
        return np.random.RandomState(random_state)

    def update(self, grad, X, y, w):
        for xb, tb in self._batch_iterator(X, y):
            w -= self.eta * np.mean(grad(xb, tb, w), axis=0)
        return w

    def _batch_iterator(self, X, y):
        idx = np.arange(X.shape[0])
        self.rgen.shuffle(idx)

        for i in range(0, X.shape[0], self.batch_size):
            begin, end = i, min(i + self.batch_size, X.shape[0])
            yield X[idx[begin:end]], y[idx[begin:end]]

class SAG(BaseOptimizer):
    def __init__(self, eta, random_state=1):
        self.eta = eta
        self.rgen = self._setup_rgen(random_state)

    @staticmethod
    def _setup_grads(grad, X, y, w):
        grads = grad(X, y, w)
        return grads

    @staticmethod
    def _setup_rgen(random_state):
        return np.random.RandomState(random_state)

    def update(self, grad, X, y, w):
        if not hasattr(self, "grads_"):
            self.grads_ = self._setup_grads(grad, X, y, w)
        k = self.rgen.randint(0, X.shape[0])
        self.grads_[k] = grad(X[k], y[k], w)
        w -= self.eta * np.mean(self.grads_, axis=0)
        return w


from abc import abstractmethod
import numpy as np

from ..base import BaseEstimator
from ..tree.tree import DecisionTreeClassifier, DecisionTreeRegressor
from ..utils.validations import check_random_state

def _get_n_sampes_bootstrap(n_samples, max_samples):
    if max_samples is None:
        return n_samples
    elif isinstance(max_samples, int):
        if not (1 <= max_samples <= n_samples):
            raise ValueError(
                f"max_samples must be in range 1 to {n_samples}, got {max_samples}"
            )
        return max_samples
    elif isinstance(max_samples, float):
        if not (0 < max_samples <= 1):
            raise ValueError(
                f"max_samples must be in range (0.0, 1.0], got {max_samples}"
            )
        return round(n_samples * max_samples)

class BaseForest(BaseEstimator):
    """
    Base class for random forest estimators.

    Parameters
    ----------
    base_estimator : BaseEstimator instance
        DecisionTreeClassifier in case of classification
        DecisionTreeRegressor in case of regression
    estimator_kwargs : dict
        keyword arguments for estimator
    n_estimators : int (default: 10)
        number of estimators
    bootstrap : bool (default: False)
        if True train estimators on random samples with replacement
        else train all estimators on whole data
    max_samples : int, float or None (default: None)
        - int: number of samples in bootstrap
        - float: fraction of whole data in bootstrap
        - None: number of samples in bootstrap = number of samples in data
    """
    def __init__(
        self,
        base_estimator,
        estimator_kwargs,
        n_estimators=10,
        bootstrap=False,
        max_samples=None,
        random_state=0
    ):
        self.base_estimator = base_estimator
        self.n_estimators = n_estimators
        self.bootstrap = bootstrap
        self.max_samples = max_samples
        self.rgen = check_random_state(random_state)
        self.estimator_kwargs = estimator_kwargs

    def _fit(self, X, y):
        trees = [
            self.base_estimator(**self.estimator_kwargs)
            for _ in range(self.n_estimators)
        ]

        self.estimators_ = []

        n_samples_bootstrap = _get_n_sampes_bootstrap(self.n_samples, self.max_samples)
        for tree in trees:
            if self.bootstrap:
                sample_indices = self.rgen.randint(0, self.n_samples, n_samples_bootstrap)
                sample_counts = np.bincount(sample_indices, minlength=self.n_samples)
                tree.fit(X, y, sample_weight=sample_counts)
            else:
                tree.fit(X, y)

        self.estimators_.extend(trees)

    @abstractmethod
    def _predict(self, X):
        raise NotImplementedError()

class RandomForestClassifier(BaseForest):
    """
    Implemntation of random forest classifier

    Parameters
    ----------
    n_estimators : int (default: 10)
        number of trees
    criterion : ["gini", "entropy"] (default: "gini")
        criterion for DecisionTreeClassifier
    max_depth : int or inf (default: inf)
        max depth of trees
    min_samples_leaf : int (default: 1)
        min samples in leaf to split node
    max_features : "sqrt", int or None (default: "sqrt")
        features to consider on each split
    bootstrap : bool (default: True)
        use bootstrap sample selection
    max_samples : int, float or None (default: None)
        - int: number of samples in bootstrap
        - float: fraction of whole data in bootstrap
        - None: number of samples in bootstrap = number of samples in data
    """
    def __init__(
        self,
        n_estimators=10,
        *,
        criterion="gini",
        max_depth=np.inf,
        min_samples_leaf=1,
        max_features="sqrt",
        bootstrap=True,
        max_samples=None,
        random_state=0
    ):
        estimator_kwargs={
            "criterion": criterion,
            "max_depth": max_depth,
            "min_samples_leaf": min_samples_leaf,
            "max_features": max_features,
            "random_state": random_state
        }
        super().__init__(
            base_estimator=DecisionTreeClassifier,
            estimator_kwargs=estimator_kwargs,
            n_estimators=n_estimators,
            bootstrap=bootstrap,
            max_samples=max_samples,
            random_state=random_state
        )

    def _fit(self, X, y):
        self.classes_ = np.unique(y)
        self.n_classes_ = self.classes_.shape[0]
        super()._fit(X, y)
    
    def _predict(self, X):
        proba = self.predict_proba(X)
        return self.classes_.take(np.argmax(proba, axis=1), axis=0)

    def predict_proba(self, X):
        all_proba = np.mean(
            [e.predict_proba(X) for e in self.estimators_],
            axis=0
        )
        return all_proba

class RandomForestRegressor(BaseForest):
    """
    Implemntation of random forest classifier

    Parameters
    ----------
    n_estimators : int (default: 10)
        number of trees
    criterion : ["mse", "mae"] (default: "mse")
        criterion for DecisionTreeClassifier
    max_depth : int or inf (default: inf)
        max depth of trees
    min_samples_leaf : int (default: 1)
        min samples in leaf to split node
    max_features : "sqrt", int or None (default: None)
        features to consider on each split
    bootstrap : bool (default: True)
        use bootstrap sample selection
    max_samples : int, float or None (default: None)
        - int: number of samples in bootstrap
        - float: fraction of whole data in bootstrap
        - None: number of samples in bootstrap = number of samples in data
    """
    def __init__(
        self,
        n_estimators=10,
        *,
        criterion="mse",
        max_depth=np.inf,
        min_samples_leaf=1,
        max_features=None,
        bootstrap=True,
        max_samples=None,
        random_state=0
    ):
        estimator_kwargs={
            "criterion": criterion,
            "max_depth": max_depth,
            "min_samples_leaf": min_samples_leaf,
            "max_features": max_features,
            "random_state": random_state
        }
        super().__init__(
            base_estimator=DecisionTreeRegressor,
            estimator_kwargs=estimator_kwargs,
            n_estimators=n_estimators,
            bootstrap=bootstrap,
            max_samples=max_samples,
            random_state=random_state
        )

    def _fit(self, X, y):
        return super()._fit(X, y)

    def _predict(self, X):
        y_hat = np.mean(
            [e.predict(X) for e in self.estimators_],
            axis=0
        )
        return y_hat


import numpy as np

def _num_samples(x):
    message = "Expected sequence or array-like, got %s" %type(x)

    if not hasattr(x, "__len__") and not hasattr(x, "shape"):
        if hasattr(x, "__array__"):
            x = np.asarray(x)
        else:
            raise TypeError(message)

    if hasattr(x, "shape") and x.shape is not None:
        if len(x.shape) == 0:
            raise TypeError("Singleton array %r cannot be considered a valid collection." % x)

        if isinstance(x.shape[0], int):
            return x.shape[0]

    try:
        return len(x)
    except TypeError:
        raise TypeError(message)

def check_consistent_length(*arrays):
    lengths = [_num_samples(X) for X in arrays if X is not None]
    uniques = np.unique(lengths)
    if len(uniques) > 1:
        raise ValueError(
            "Found input variables with inconsistent numbers of samples: %r"
            % [int(l) for l in lengths]
        )

def check_random_state(seed):
    if seed is None or seed is np.random:
        return np.random.mtrand._rand
    if isinstance(seed, int):
        return np.random.RandomState(seed)
    if isinstance(seed, np.random.RandomState):
        return seed
    raise ValueError(
        f"{seed} can not be used for RandomState instance"
    )

def check_sample_weight(sample_weight, X):
    n_samples = _num_samples(X)

    if sample_weight is None:
        sample_weight = np.ones(n_samples, dtype=np.float64)
    elif isinstance(sample_weight, (int, float)):
        sample_weight = np.full(n_samples, sample_weight, dtype=np.float64)
    else:
        if not isinstance(sample_weight, np.ndarray):
            sample_weight = np.asarray(sample_weight)
        if sample_weight.ndim != 1:
            raise ValueError(
                f"sample_weights must be one dimensional array-like, got{sample_weight.ndim}"
            )
        if sample_weight.shape != (n_samples,):
            raise ValueError(
                f"sample_weights shape must be ({n_samples},), got {sample_weight.shape}"
            )

    return sample_weight


import numpy as np

from ..base import BaseEstimator
from .kd_tree import KDTree
from .ball_tree import BallTree

algorithms = ["brute", "kd-tree", "ball-tree"]
metrics = {"l-1" : 1, "l-2" : 2, "l-inf" : np.inf}
weights_modes = ["uniform", "distance", "gaussian"]

class KNNBase(BaseEstimator):
    """
    Implementation of k Nearest Neighbors algorithm

    Attributes
    ----------
    k : int (default: 5)
        number of neighbors
    algorithm : str (default: "kd-tree")
        algorithm of neighbor search
        - "brute" : brute-force approach (no fit required)
        - "kd-tree" : kd-tree approach (fit required to build tree)
        - "ball-tree" : ball-tree approach (fit required to build tree)
        In case of tree algorithms using simple depth-first search.
        See `_query_single_depthfirst` in `sklearn/neighbors/_binary_tree.pxi`
    metric : str (default: "l-2")
        distance metric
        - "l-2": euclidian distance
        - "l-1": manhattan distance
        - "l-inf": max-abs distance
    weights-mode : str (default: "uniform")
        weights in decision based on distance
        - "uniform" : weights independent from distances
        - "distance" : weights = 1 / distances
        - "gaussian" : weights = exp(-2 * distances^2) / sqrt(2 * pi)
    """
    def __init__(self, k=5, algorithm="kd-tree", metric="l-2", weights_mode="uniform"):
        if k < 1:
            raise ValueError("k must be greater than 0.")
        if algorithm not in algorithms:
            raise ValueError(f"Unknown algorithm: availible only {algorithms}")
        if metric not in metrics:
            raise ValueError(f"Unknown metric: availible only {list(metrics.keys())}")
        if weights_mode not in weights_modes:
            raise ValueError(f"Unknown weights mode: availible only {weights_modes}")
        self.k = k
        self.algorithm = algorithm
        self.metric = metrics[metric]
        self.weights_mode = weights_mode

    def _fit(self, X, y):
        self.X = X
        self.y = y
        if self.algorithm == "kd-tree":
            self.tree = KDTree(X, y)
        elif self.algorithm == "ball-tree":
            self.tree = BallTree(X, y)
        elif self.algorithm == "brute":
            pass

    def _get_weights(self, distances):
        if self.weights_mode == "uniform":
            return None
        elif self.weights_mode == "distance":
            return 1 / distances
        elif self.weights_mode == "gaussian":
            return np.exp(-2 * distances**2) / np.sqrt(2 * np.pi)

    def _aggregate(self, neighbors_targets, weights):
        raise NotImplementedError()

    def _predict(self, X):
        return np.array([self._predict_x(x) for x in X])

    def _predict_x(self, x):
        if self.algorithm == "kd-tree":
            neighbors_targets, distances = self.tree.query_knn(x, self.k, metric=self.metric)
            weights = self._get_weights(distances)
        elif self.algorithm == "ball-tree":
            neighbors_targets, distances = self.tree.query_knn(x, self.k, metric=self.metric)
            weights = self._get_weights(distances)
        elif self.algorithm == "brute":
            distances = np.array([np.linalg.norm(x - example) for example in self.X])
            idx = np.argsort(distances)
            neighbors_targets = self.y[idx][:self.k]
            weights = self._get_weights(distances[idx][:self.k])
        
        return self._aggregate(neighbors_targets, weights)
        
class KNNClassifier(KNNBase):
    """
    Multinominal classification with KNN.
    Using np.bincount to get number of nearest class samples.
    """
    def _fit(self, X, y):
        self.classes_ = np.unique(y)
        self.n_classes_ = self.classes_.shape[0]
        return super()._fit(X, y)

    def _aggregate(self, neighbors_targets, weights):
        labels = np.bincount(neighbors_targets, weights=weights)
        return self.classes_.take(labels.argmax(), axis=0)


class KNNRegressor(KNNBase):
    """
    Regression with KNN.
    Using weighted average.
    """
    def _aggregate(self, neighbors_targets, weights):
        return np.average(neighbors_targets, weights=weights)


from mlhandmade import (
    linear_models,
    discriminant_analysis,
    knn,
    kernel_methods,
    tree,
    ensemble,
    multiclass,
    preprocessing,
    plotting,
    model_selection
)

__all__ = [
    "linear_models",
    "discriminant_analysis",
    "knn",
    "kernel_methods",
    "tree",
    "ensemble",
    "multiclass",
    "preprocessing",
    "plotting",
    "model_selection"
]


import numpy as np

from ..base import BaseEstimator
from .linear import LinearKernel
from .poly import PolynomialKernel
from .rbf import RBFKernel

kernels = {
    "linear" : LinearKernel,
    "poly" : PolynomialKernel,
    "rbf" : RBFKernel
}

class BaseSVM(BaseEstimator):
    """
    Base class for SVM implements QP solver

    Parameters
    ----------
    C : float
        constraint for lagrange multipliers
        in dual problem
    """
    def __init__(self, C):
        self.C = C
    
    def _solve_qp(self, coef, grad, t, Q, tol):
        """
        Solve QP problem with one linear constraint [1]:

        min ->coef | 0.5 * coef^T @ Q @ coef - grad^T @ coef
        subject to | t^T @ coef = 0
                   | 0 <= coef[i] <= C, i = 1, ..., N

        Reference:
        [1] `Chang and Lin, LIBSVM: A Library for Support Vector Machines.`.
        [2] ctgk/PRML/prml/kernel on github.

        Parameters
        ----------
        X : (N, D) np.ndarray
            training independent variable
        t : (N,) np.ndarray
            training dependent variable
            binary -1 or 1
        tol : float, optional
            numerical tolerance (the default is 1e-8)
        """
        while True:
            # Select working set with WSS 1 "maximal violating pair"[1]
            # i = argmax{-y * grad | y from I_up}
            # j = argmin{-y * grad | y from I_low}, where
            # I_up = {y | (coef < C) & (y = 1) or (coef > 0) & (y = -1)}
            # I_low = {y | (coef < C) & (y = -1) or (coef > 0) & (y = 1)}
            tg = t * grad
            mask_up = (t == 1) & (coef < self.C - tol)
            mask_up |= (t == -1) & (coef > tol)
            mask_down = (t == -1) & (coef < self.C - tol)
            mask_down |= (t == 1) & (coef > tol)
            i = np.where(mask_up)[0][np.argmax(tg[mask_up])]
            j = np.where(mask_down)[0][np.argmin(tg[mask_down])]
            
            # check convergence
            if tg[i] < tg[j] + tol:
                # b must satisfy conditions tg[i] < b < tg[j],
                # so it's no matter how we sum tg's, average is a good choice[1]
                self.b = 0.5 * (tg[i] + tg[j])
                break
            # Solve two-variable problem
            # A, B - constraints: (a_i, a_j) in the box [0, C] x [0, C]
            A = self.C - coef[i] if t[i] == 1 else coef[i]
            B = coef[j] if t[j] == 1 else self.C - coef[j]
            direction = (tg[i] - tg[j]) / (Q[i, i] - 2 * Q[i, j] + Q[j, j])
            # limit coef's if direction is too large
            direction = min(A, B, direction)
            coef[i] += direction * t[i]
            coef[j] -= direction * t[j]
            grad -= direction * t * (Q[i] - Q[j])
        
        return coef

    def decision_function(self, X):
        """
        SVM classifiers can be used with multiclass strategies
        """
        raise NotImplementedError()

class SupportVectorClassifier(BaseSVM):
    """
    support vector classifier algorithm requires solution of 
    the next QP problem:

    min ->coef | 0.5 * coef^T @ Q @ coef - grad^T @ coef
    subject to | t^T @ coef = 0
               | 0 <= coef[i] <= C, i = 1, ..., N

    where Q = Gram, Gram = K(X, X)
    grad - ones(N)
    coef - zeros(N)
    t - class labels y

    Parameters
    ----------
    C : float (default: np.inf)
        inverse regularization parameter (inf - no penalty)
    kernel : str (default: "linear")
        kernel insted of inner product
    kernel_kwargs : kwargs
        parameters of kernel
    """
    def __init__(self, C=np.inf, kernel="linear", **kernel_kwargs):
        if kernel not in kernels:
            raise ValueError(f"Unkown kernel: avilible only {list(kernels.keys())}")
        self.kernel = kernels[kernel](**kernel_kwargs)
        super().__init__(C)

    def _fit(self, X, y, tol=1e-8):
        """
        estimate support vectors and their parameters
        Parameters
        ----------
        X : (N, D) np.ndarray
            training independent variable
        y : (N,) np.ndarray
            training dependent variable
            binary -1 or 1
        tol : float, optional
            numerical tolerance (the default is 1e-8)
        """
        N = len(y)
        coef = np.zeros(N) # alphas(lagrangian multipliers)
        grad = np.ones(N)
        t = y
        Q = self.kernel(X, X)
        alpha = super()._solve_qp(coef, grad, t, Q, tol)
        support_mask = alpha > tol
        self.a = alpha[support_mask]
        self.X = X[support_mask]
        self.y = y[support_mask]

    def _predict(self, X):
        """
        predict labels of the input

        Parameters
        ----------
        x : (sample_size, n_features) ndarray
            input

        Returns
        -------
        label : (sample_size,) ndarray
            predicted labels
        """
        y = self.decision_function(X)
        labels = np.sign(y)
        return labels

    def decision_function(self, X):
        """
        calculate distance from the decision boundary

        Parameters
        ----------
        x : (sample_size, n_features) ndarray
            input

        Returns
        -------
        distance : (sample_size,) ndarray
            distance from the boundary
        """
        distance = np.sum(
            self.a * self.y
            * self.kernel(X, self.X),
            axis=-1) + self.b
        return distance

class SupportVectorRegressor(BaseSVM):
    """
    support vector classifier algorithm requires solution of 
    the next QP problem:

    min ->coef | 0.5 * coef^T @ Q @ coef - grad^T @ coef
    subject to | t^T @ coef = 0
               | 0 <= coef[i] <= C, i = 1, ..., N

    where Q = [[Gram, -Gram]
               [-Gram, Gram]], Gram = K(X, X)
    grad = - [eps * ones(N) - y, eps * ones(N) + y]
    coef = zeros(2 * N)
    t = [ones(N), -ones(N)]

    Parameters
    ----------
    C : float (default: np.inf)
        inverse regularization parameter (inf - no penalty)
    kernel : str (default: "linear")
        kernel insted of inner product
    kernel_kwargs : kwargs
        parameters of kernel
    """
    def __init__(self, C=np.inf, eps=0.1, kernel="linear", **kernel_kwargs):
        if kernel not in kernels:
            raise ValueError(f"Unkown kernel: avilible only {list(kernels.keys())}")
        self.kernel = kernels[kernel](**kernel_kwargs)
        self.eps = eps
        super().__init__(C)

    def _fit(self, X, y, tol=1e-8):
        """
        estimate support vectors and their parameters
        Parameters
        ----------
        X : (N, D) np.ndarray
            training independent variable
        y : (N,) np.ndarray
            training dependent variable
            binary -1 or 1
        tol : float, optional
            numerical tolerance (the default is 1e-8)
        """
        N = len(y)
        coef = np.zeros(2 * N) # alphas(lagrangian multipliers)
        grad = -np.r_[self.eps * np.ones(N) - y, self.eps * np.ones(N) + y]
        t = np.r_[np.ones(N), -np.ones(N)]
        Gram = self.kernel(X, X)
        Q = np.c_[np.r_[Gram, -Gram], np.r_[-Gram, Gram]]
        alpha = super()._solve_qp(coef, grad, t, Q, tol)
        alpha = alpha[:N] - alpha[N:]
        support_mask = np.abs(alpha) > tol
        self.a = alpha[support_mask]
        self.X = X[support_mask]
        self.y = y[support_mask]

    def _predict(self, X):
        """
        predict values of the input

        Parameters
        ----------
        x : (sample_size, n_features) ndarray
            input

        Returns
        -------
        label : (sample_size,) ndarray
            predicted values
        """
        y = self.decision_function(X)
        return y

    def decision_function(self, X):
        """
        calculate target value

        Parameters
        ----------
        x : (sample_size, n_features) ndarray
            input

        Returns
        -------
        distance : (sample_size,) ndarray
            target value
        """
        target = np.sum(
            self.a
            * self.kernel(X, self.X),
            axis=-1) + self.b
        return target


import numpy as np
import warnings

class BaseCriterion:
    def gain(self, y, splits, sample_weight):
        n = y.size
        n_left, n_right = splits[0].size, splits[1].size
        left, right = self(splits[0], sample_weight[0]), self(splits[1], sample_weight[1])
        return n_left / n * left + n_right / n * right

class ClassificationCriterion(BaseCriterion):
    def __init__(self, criterion):
        self.criterion = criterion
    
    def __call__(self, x, sample_weight):
        if self.criterion == "gini":
            return self.gini(x, sample_weight)
        elif self.criterion == "entropy":
            return self.entropy(x, sample_weight)
        else:
            raise ValueError("Got unknown criterion.")

    @staticmethod
    def entropy(x, sample_weight):
        # log2(p) raise a RuntimeWarning even with np.where()
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            p = np.bincount(x, weights=sample_weight) / np.sum(sample_weight)
            entrs = np.where(p != 0.0, -p * np.log2(p), 0.0)
            return np.sum(entrs)

    @staticmethod
    def gini(x, sample_weight):
        p = np.bincount(x, weights=sample_weight) / np.sum(sample_weight)
        return 1 - np.sum(p**2)

class RegressionCriterion(BaseCriterion):
    def __init__(self, criterion):
        self.criterion = criterion

    def __call__(self, x, sample_weight):
        if self.criterion == "mse":
            return self.mse(x, sample_weight)
        elif self.criterion == "mae":
            return self.mae(x, sample_weight)
        else:
            raise ValueError("Got unknown criterion.")

    @staticmethod
    def mse(x, sample_weight):
        x_mean = np.average(x, weights=sample_weight)
        return np.average((x - x_mean)**2, weights=sample_weight)
    
    @staticmethod
    def mae(x, sample_weight):
        if sample_weight is None:
            x_median = np.median(x)
        else:
            x_median = np.median(x * sample_weight)
        return np.average(np.abs(x - x_median), weights=sample_weight)


from mlhandmade.preprocessing.data_preprocessing import (
    data_shuffle,
    ordinal,
    binary,
    add_bias_feature,
    standardize,
    onehot,
    polynomial_features
)

__all__ = [
    "data_shuffle",
    "ordinal",
    "binary",
    "add_bias_feature",
    "standardize",
    "onehot",
    "polynomial_features"
]


import numpy as np
from .kernel import Kernel

class RBFKernel(Kernel):
    """
    Radial basis function kernel:
    k(x, y) = C0 * exp(-0.5 * (x - y).T @ P @ (x - y)),
    where P is a diagonal matrix of parameters.
    with `P = I / sigma^2` and `C0 = 1` we get classical
    representation from C. Bishop's PRML book

    Attributes
    ----------
    params : ndarray of shape (ndim + 1,)
        parameters of radial basis function
    ndim : int
        dimension of expected input data
    """
    def __init__(self, params):
        """
        construct Radial basis kernel function
        """
        if not isinstance(params, np.ndarray):
            params = np.asarray(params)
        assert params.ndim == 1
        self.params = params
        self.ndim = len(params) - 1

    def __call__(self, x, y, pairwise=True):
        """
        calculate radial basis function

        Parameters
        ----------
        x : ndarray [..., ndim]
            input of this kernel function
        y : ndarray [..., ndim]
            another input

        Returns
        -------
        output : ndarray
            output of this radial basis function
        """
        assert x.shape[-1] == self.ndim
        assert y.shape[-1] == self.ndim
        if pairwise:
            x, y = self._pairwise(x, y)
        d = self.params[1:] * (x - y) ** 2
        return self.params[0] * np.exp(-0.5 * np.sum(d, axis=-1))

    def derivatives(self, x, y, pairwise=True):
        if pairwise:
            x, y = self._pairwise(x, y)
        d = self.params[1:] * (x - y) ** 2
        delta = np.exp(-0.5 * np.sum(d, axis=-1))
        deltas = -0.5 * (x - y) ** 2 * (delta * self.params[0])[:, :, np.newaxis]
        return np.concatenate((np.expand_dims(delta, 0), deltas.T))

    def update_parameters(self, updates):
        self.params += updates


from mlhandmade.linear_models.linear_classifier import (
    LinearClassifier,
    SoftmaxRegressor
)

from mlhandmade.linear_models.linear_regressor import (
    LinearRegressor,
    RidgeRegressor
)

__all__ = [
    "LinearClassifier",
    "SoftmaxRegressor",
    "LinearRegressor",
    "RidgeRegressor"
]


import numpy as np
from itertools import product
from functools import partial, reduce
from operator import mul

from ..base import BaseEstimator
from ..model_selection import cross_val_score
from ..utils.validations import check_random_state

class ParamGrid:
    def __init__(self, param_grid):
        if not isinstance(param_grid, (dict, list)):
            raise ValueError(
                f"`params` must be a dict or a list, but got {type(param_grid).__name__}."
            )
        if isinstance(param_grid, dict):
            param_grid = [param_grid]
        self.param_grid = param_grid

    def __iter__(self):
        for p in self.param_grid:
            items = sorted(p.items())
            if not items:
                yield {}
            else:
                keys, values = zip(*items)
                for v in product(*values):
                    params = dict(zip(keys, v))
                    yield params

    def __len__(self):
        product = partial(reduce, mul)
        return sum(
            product(len(v) for v in p.values()) if p else 1 for p in self.param_grid
        )

class GridSearchCV(BaseEstimator):
    def __init__(
        self,
        estimator,
        param_grid,
        *,
        scoring=None,
        cv=5,
        shuffle=False,
        random_state=0,
        **score_kwargs
    ):
        self.estimator = estimator
        self.param_grid = ParamGrid(param_grid)
        self.scoring = scoring
        self.cv = cv
        self.shuffle = shuffle
        self.rgen = check_random_state(random_state)
        self.score_kwargs = score_kwargs
    
    def _fit(self, X, y):
        estimators = []
        params = []
        scores = np.zeros(len(self.param_grid))
        for i, param in enumerate(self.param_grid):
            estimator = self.estimator(**param)
            score = cross_val_score(
                estimator,
                X,
                y,
                score=self.scoring,
                cv=self.cv,
                shuffle=self.shuffle,
                random_state=self.rgen,
                **self.score_kwargs
            )
            estimators.append(estimator)
            params.append(param)
            scores[i] = score

        best_idx = np.argmax(scores)
        self.best_estimator_ = estimators[best_idx]
        self.best_params_ = params[best_idx]
        self.best_score_ = scores[best_idx]
        self.best_estimator_.fit(X, y)

    def _predict(self, X):
        return self.best_estimator_.predict(X)


from setuptools import setup, find_packages
from codecs import open
from os import path

__version__ = "0.0.12"

here = path.abspath(path.dirname(__file__))

with open(path.join(here, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="mlhandmade",
    version=__version__,
    description="Implemented some ML routines",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="gittasche",
    python_reqires=">=3.7",
    install_requires=["numpy", "matplotlib"],
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False
)


import numpy as np
from numpy.typing import ArrayLike

from .base import KDTreeBase

class KDTree(KDTreeBase):
    """
    Numpy based kd-tree.

    Parameters
    ----------
    points : array-like of shape (N, D)
        Array of points to build tree
    values : array-like of shape(N,)
        values of some function in points, 
        for instance binary classes or continuos value in case of regression
    """
    def __init__(self, points: ArrayLike, values: ArrayLike) -> None:
        if not isinstance(points, np.ndarray):
            points = np.array(points)

        self.points = np.atleast_2d(points)
        super().__init__(self.points, values)
    
    def query_knn(
        self,
        point: ArrayLike,
        k: int,
        sort: bool = False,
        metric: int = 2
    ) -> np.ndarray:
        """
        k nearest neighbors query.

        Parameters
        ----------
        point : array-like of shape (D,)
            get neighbors of this point
        k : int
            number of neighbors
        sort : bool (default: False)
            sort neighbors by distance to point
        metric : {1, 2, numpy.inf}
            metric of distance for numpy.linalg.norm function.
            1 : l1 norm, sum(abs(x[i]))
            2 : l2 norm, sqrt(sum(x[i]**2))
            numpy.inf : l-inf norm, max(abs(x[i]))

        Returns
        -------
        ndarray of ndarray's of neighbors and distances if return_distances
        """
        if not isinstance(point, np.ndarray):
            point = np.array(point)
        
        if k < 1:
            raise ValueError("k must be greater than 0.")
        
        if point.shape[0] != super().get_dim:
            raise ValueError(f"Point dimension {point.shape[1]} \
             must equal tree dimension {super().get_dim}")

        heap = []
        super()._query_knn(point, k, metric, heap)
        if sort:
            heap = sorted(heap)[::-1]
        
        values = np.array([h[3] for h in heap])
        distances = np.array([-h[0] for h in heap])
        
        return values, distances


import numpy as np
import matplotlib.pyplot as plt
from itertools import product

from ..model_selection import confusion_matrix

def plot_confusion_matrix(
    y_true,
    y_pred,
    *,
    include_values=True,
    cmap="viridis",
    xticks_rotation="horizontal",
    values_format=None,
    display_labels=None,
    ax=None,
    colorbar=True,
    im_kw=None,
    **cm_kwargs
):
    if ax is None:
        fig, ax = plt.subplots()
    else:
        fig = ax.figure
    
    cm = confusion_matrix(y_true, y_pred, **cm_kwargs)
    n_classes = cm.shape[0]
    
    default_im_kw = dict(interpolation="nearest", cmap=cmap)
    im_kw = im_kw or {}
    im_kw = {**default_im_kw, **im_kw}

    im = ax.imshow(cm, **im_kw)
    text = None
    cmap_min, cmap_max = im.cmap(0), im.cmap(1.0)

    if include_values:
        text = np.empty_like(cm, dtype=object)

        # `text` color is depends from color of background
        thresh = (cm.max() + cm.min()) / 2.0

        # cartesian product of two axes of confusion matrix
        for i, j in product(range(n_classes), range(n_classes)):
            color = cmap_max if cm[i, j] < thresh else cmap_min

            if values_format is None:
                text_cm = format(cm[i, j], ".2g")
                if cm.dtype.kind != "f":
                    text_d = format(cm[i, j], "d")
                    if len(text_d) < len(text_cm):
                        text_cm = text_d
            else:
                text_cm = format(cm[i, j], values_format)
            
            text[i, j] = ax.text(
                j, i, text_cm, ha="center", va="center", color=color
            )

    if display_labels is None:
        display_labels = np.arange(n_classes)
    
    if colorbar:
        fig.colorbar(im, ax=ax)
    ax.set(
        xticks=np.arange(n_classes),
        yticks=np.arange(n_classes),
        xticklabels=display_labels,
        yticklabels=display_labels,
        ylabel="True label",
        xlabel="Predicted label"
    )

    ax.set_ylim((n_classes - 0.5, -0.5))
    plt.setp(ax.get_xticklabels(), rotation=xticks_rotation)

    return ax


import numpy as np
import heapq
import itertools

class BallTreeBase:
    """
    Implemnatation of ball-tree class.
    This implementation similar to sklearn binary tree,
    but with `leaf_size` = 1.
    Ball-tree contains more nodes than kd-tree, so it is a
    little bit slower, but still faster than brute force
    approach.

    Attributes:
    points : ndarray of shape (N, D)
        points of train dataset
    values : ndarray of shape(N,)
        values of some function in `points`,
        for example, continuos values in case of regression
    metric : int (default: 2)
        `ord` for `np.linalg.norm`
    """
    def __init__(self, points, values, metric=2):
        if points.shape[0] > 1:
            best_dimension = np.argmax(np.max(points, axis=0) - np.min(points, axis=0))
            idx_sort = np.argsort(points[:, best_dimension])
            points = points[idx_sort]
            values = values[idx_sort]

            middle = points.shape[0] >> 1
            self.data = np.mean(points, axis=0)
            self.value = values[middle]
            self.radius = np.max(np.linalg.norm(self.data - points, ord=metric, axis=-1))
            self.left = BallTreeBase(points[:middle], values[:middle], metric)
            self.right = BallTreeBase(points[middle:], values[middle:], metric)
        elif points.shape[0] == 1:
            self.data = points[0]
            self.value = values[0]
            self.left = None
            self.right = None
            self.radius = 0

    def _query_knn(self, point, k, metric, heap):
        """
        Function computes reduced distance between root and point
        RDist(node.data, point) = Dist(node.data, point) - node.radius
        """
        dist = self.distance(self.data, point, metric) - self.radius
        self._query_knn_reduced(point, k, metric, heap, dist)

    def _query_knn_reduced(self, point, k, metric, heap, rdist, counter=itertools.count()):
        if len(heap) and rdist >= -heap[0][0]:
            return
        elif self.left is None and self.right is None:
            # Case of leaf node: Dist(leaf.data, point) = RDist(leaf.data, point)
            item = (-rdist, next(counter), self.data, self.value)
            if len(heap) < k:
                heapq.heappush(heap, item)
            elif rdist < -heap[0][0]:
                heapq.heappushpop(heap, item)
        else:
            left_rdist = self.distance(self.left.data, point, metric) - self.left.radius
            right_rdist = self.distance(self.right.data, point, metric) - self.right.radius
            if left_rdist <= right_rdist:
                self.left._query_knn_reduced(point, k, metric, heap, left_rdist, counter)
                self.right._query_knn_reduced(point, k, metric, heap, right_rdist, counter)
            else:
                self.right._query_knn_reduced(point, k, metric, heap, right_rdist, counter)
                self.left._query_knn_reduced(point, k, metric, heap, left_rdist, counter)
        return

    @staticmethod
    def distance(point1, point2, metric):
        return np.linalg.norm(point1 - point2, ord=metric)